#!/usr/bin/env python3
#-*- coding:utf-8 -*-
##
## main.py
##
##  Created on: Fev 20, 2020
##      Author: Antonio Morgado
##      E-mail: ajrmorgado@gmail.com
##
# Copyright (C) 2020, Antonio Morgado, Mikolas Janota
#
# This file is part of dtfinder.
#
# dtfinder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dtfinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dtfinder.  If not, see <https://www.gnu.org/licenses/>.

#
#==============================================================================
#
# required imports

import sys
import getopt
import os
import inspect
import math
import time
from pysat.solvers import Solver
from pysat.formula import IDPool , CNF, WCNF
from pysat.card import *
from pysat.examples.rc2 import RC2

import mytrie

sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../scripts/'))
import enum_topologies

import functools
print = functools.partial(print, flush=True)


sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../greedy/'))

import run_sklearn

import dtfinderTopDT1
import dtfinderTopPathsEncd

#
#==============================================================================
#
def preprocess(samples, names, quasipure, verbosity):
    
    k = len(samples[0])-1
    
    fcount = [[[0,0],[0,0]] for x in range(k)]
    
    for i in range(len(samples)):
        acpt = samples[i][k]
        for j in range(k):
            obs = samples[i][j]
            fcount[j][obs][acpt] = fcount[j][obs][acpt] + 1
    
    onlyOneValue = set()
    
    for j in range(k):
        if fcount[j][0][0] == 0 and fcount[j][0][1] == 0:
            onlyOneValue.add(j)
        elif fcount[j][1][0] == 0 and fcount[j][1][1] == 0:
            onlyOneValue.add(j)
            
    newsamples = []
    newnames = []
    
    if quasipure is not None:
        quasipure = {}
    
    for j in range(k):
        if j not in onlyOneValue:
            newnames.append(names[j])
    newnames.append(names[k])
    
    for s in samples:
        ns = []
        for j in range(k):
            if  j not in onlyOneValue:
                ns.append(s[j])
                if quasipure is not None:
                    if fcount[j][0][0] == 0 or fcount[j][0][1] == 0 or fcount[j][1][0] == 0 or fcount[j][1][1] == 0:
                        index = len(ns)-1
                        quasipure[index] = fcount[j]
        ns.append(s[k])
        newsamples.append(ns)
        
    if verbosity > 0:
        print('# prepocess number of removed features:', len(onlyOneValue))
        if quasipure is not None:
            print('# prepocess number of quasipure features:', len(quasipure))

    return newsamples, newnames, quasipure
#
#==============================================================================
# read an example line
def read_example_line(line, examples, verbosity):
    sample = line.split(',')
    for i in range(len(sample)):
        sample[i] = int(sample[i])
    examples.append(sample)
#
#==============================================================================
# parser for the examples
def read_examples_file(filepath, verbosity):

    if not os.path.isfile(filepath):
        print("File {} does not exist. Does it?".format(filepath))
        sys.exit()
        
    examples = []
    with open(filepath) as fp:
        line = fp.readline()
        featNames = []
        if line:
            featNames = line.strip().split(',')
            line = fp.readline()
        else:
            return
        while line:
            line = line.strip()
            if len(line) > 0 :
                read_example_line(line.strip(), examples, verbosity)
            line = fp.readline()
    return examples, featNames
#
#==============================================================================
#
def main(argv):
    algorithm = 'lssu'
    pbounds = 0
    cencd = 0
    din = 1
    encd = 0
    inputfile = ''
    lbencd = 0
    matchencd = 0
    npaths = -1
    nnodes = -1
    quasipure = None
    recursive = 0
    solver = 'g3'
    tpldepth = 3
    ordertopologies = 0
    ubencd = 0
    useassumps = 500
    verbosity = 0
    
    try:
        opts, args = getopt.getopt(argv,"a:b:c:d:e:hi:lm:qrs:t:o:uvx:",["algorithm", "cardenc=", "encoding=", "ifile=", "nnodes=", "npaths=",  "verbosity", "solver", "setpathbounds", "setmatchencd", "setquasipure", "setdinamic"])
    except getopt.GetoptError:
        print('Error in input parameters, for help pease use: $dtfinder.py -h')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('dtfinder.py -i <inputfile> [<options>]')
            print(' -a <str> is the algorithm, lssu (default), lsus, bs, prgsu, prgus')
            print(' -b <int> is the type of path size bounds to use')
            print('       0 -- no bounds are used')
            print('       1 -- upper bounds of maxones and maxzeros')
            print('       2 -- same as 1 plus lower bounds using maxsat')
            print('       3 -- same as 2 plus using or constraints in maxsat')
            print('       4 -- same as 3 plus upper bounds using maxsat (not yet done)')
            print(' -c <int> is the cardinality encoding number, 0 = Pairwise(default), 1 = Seq Counter, ... (any allowed by pysat)')
            print(' -e <int> is the encoding number, 0 = DT1(default), 1 = Paths')
            print(' -l encode the uppers bound of the DT1 encoding (section 3.3)')
            print(' -m <int> sets the encoding of discrimination in the Paths encoding')
            print('       0 -- no match variables are created')
            print('       1 -- creates a match variable for each path and each sample')
            print('       2 -- uses match variables with a trie of the samples')
            #print(' -n <int> is the upper bound on the number of nodes')
            #print(' -p <int> is the upper bound on the number of paths')
            print(' -s <str> is the sat solver to be used, g3(default), g4, cd, lgl, mcb, ... (any allowed by pysat)')
            print(' -t <int> topology depth if number of nodes > 21')
            print(' -o <int> if > 0 then score and sort topologies (default:0 [no order])')
            print(' -u encode the uppers bound of the DT1 encoding (section 3.3)')
            print(' -v is the verbosity')
            print(' -x <int> min number of topologies to start using assumptions')
            sys.exit()
        elif opt in ("-a", "--algorithm"):
            algorithm = arg
        elif opt in ("-b", "--setpathbounds"):
            pbounds = int(arg)
        elif opt in ("-c", "--cardenc"):
            cencd = int(arg)
        elif opt in ("-d", "--setdinamic"):
            din = int(arg)
        elif opt in ("-e"):
            encd = int(arg)
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt == "-l":
            lbencd = 1
        elif opt in ("-m", "--setmatchencd"):
            matchencd = int(arg)
        #elif opt in ("-n", "--nnodes"):
        #    nnodes = int(arg)
        #    npaths = math.floor(((nnodes+1)/2))
        #elif opt in ("-p", "--npaths"):
        #    npaths = int(arg)
        #    nnodes = 2*npaths-1
        elif opt in ("-q", "--setquasipure"):
            quasipure = {}
        elif opt == "-r":
            recursive = 1
        elif opt in ("-s", "--solver"):
            solver = arg
        elif opt in ("-t"):
            tpldepth = int(arg)
        elif opt in ("-o"):
            ordertopologies = int(arg)
        elif opt == "-u":
            ubencd = 1
        elif opt in ("-v", "--verbosity"):
            verbosity = 1
        elif opt in ("-x"):
            useassumps = int(arg)
            
    examples, featNames = read_examples_file(inputfile, verbosity)
    
    if verbosity > 0:
        print('# nfeat:{} nsamples:{}'.format(len(examples[0])-1, len(examples)))
    
    examples, featNames, quasipure = preprocess(examples, featNames, quasipure, verbosity)
    
    trie = None
    if matchencd > 1:
    
        t0 = time.time()
        trie = mytrie.Trie()
        if din > 0:
            trie.setsamplesdin(examples)
        else:
            trie.setsamples(examples)
        if verbosity > 0:
            print('# Trie building time: {:.2f}'.format(time.time()-t0))
            print('# n.nodes in trie: {} (compression rate: {:.2f})'.format(trie.nid, (trie.nid / ((len(examples[0])-1)* len(examples)) ) ))
   
    nnodes, SKL = run_sklearn.run_sk_learn(examples)
    npaths = int((nnodes+1)/2)
        
    SKL_TOPO = enum_topologies.sklearn2topology(SKL)
    enum_topologies.number_tree(SKL_TOPO)
        
    if verbosity > 0:
        print('# ub-paths:', npaths, 'ub-nodes:', nnodes)
        print ("# sklearn topology:", end = '')
        SKL_TOPO.print()
        print()
        #print('-----')
        #enum_topologies.show_nfolhas(SKL_TOPO)
        #print('-----')
    
    if encd == 0 :
        solver = dtfinderTopDT1.DT1(examples, solver, algorithm, featNames, cencd, nnodes, ubencd, lbencd, quasipure, matchencd, trie, SKL, SKL_TOPO, tpldepth, ordertopologies, useassumps, verbosity)
        solver.solve()
        
    else:
        solver = dtfinderTopPathsEncd.PathsEnc(examples, featNames, solver, algorithm, cencd, npaths, pbounds, recursive, quasipure, matchencd, trie, SKL, SKL_TOPO, tpldepth, ordertopologies, useassumps, verbosity)
        solver.solve()
#
#==============================================================================

if __name__ == "__main__":
    main(sys.argv[1:])
