#!/usr/bin/env python3
#-*- coding:utf-8 -*-
##
## main.py
##
##  Created on: Fev 20, 2020
##      Author: Antonio Morgado
##      E-mail: ajrmorgado@gmail.com
##
# Copyright (C) 2020, Antonio Morgado, Mikolas Janota
#
# This file is part of dtfinder.
#
# dtfinder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dtfinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dtfinder.  If not, see <https://www.gnu.org/licenses/>.
#
#==============================================================================
#
# required imports

import sys
import getopt
import os
import inspect
import math
import time
from pysat.solvers import Solver
from pysat.formula import IDPool , CNF, WCNF
from pysat.card import *
from pysat.examples.rc2 import RC2

import mytrie

sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../scripts/'))

import enum_topologies

import functools
print = functools.partial(print, flush=True)

sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../greedy/'))

import run_sklearn

#
#==============================================================================
#
class PathsEnc(object):
    #
    #--------------------------------------------------------------------------
    def atMostOne(self, literals):
    
        encType = EncType.pairwise
        
        if len(literals) > 7:
        
            if self.card == 1:
                encType = EncType.seqcounter
            elif self.card == 2:
                encType = EncType.sortnetwrk
            elif self.card == 3:
                encType = EncType.cardnetwrk
            elif self.card == 4:
                encType = EncType.bitwise
            elif self.card == 5:
                encType = EncType.ladder
            elif self.card == 6:
                encType = EncType.totalizer
            elif self.card == 7:
                encType = EncType.mtotalizer
            elif self.card == 8:
                encType = EncType.kmtotalizer
            
        am1 = CardEnc.atmost(lits=literals, top_id=self.vpool._next(), encoding=encType)
        self.vpool.occupy(self.vpool._next(), am1.nv)
        return am1
    #
    #--------------------------------------------------------------------------
    def maxpath(self):
        return min(self.N-1, self.K)
    #
    #--------------------------------------------------------------------------
    def create_variables(self, topv):
        self.vpool = IDPool(start_from=topv+1)
        
        self.xvar = lambda i,j: self.vpool.id('x{0} {1}'.format(i,j))
        self.tvar = lambda i,j: self.vpool.id('t{0} {1}'.format(i,j))
        self.cvar = lambda i: self.vpool.id('c{0}'.format(i))
        self.evar = lambda i,j: self.vpool.id('e{0} {1}'.format(i,j))
        self.fvar = lambda i,j,k: self.vpool.id('f{0} {1} {2}'.format(i,j,k))
        self.d0var = lambda i,k: self.vpool.id('d0 {0} {1}'.format(i,k))
        self.d1var = lambda i,k: self.vpool.id('d1 {0} {1}'.format(i,k))
        
        self.auxvar = lambda : self.vpool.id('aux{}'.format(self.vpool._next()))
    #
    #--------------------------------------------------------------------------
    def print_variables(self):
        print('M={0}'.format(self.M))
        print('K={0}'.format(self.K))
        print('N={0}'.format(self.N))
        
        print('max path size:', self.maxpath())
        
        for i in range(1,self.N+1):
            for j in range(1, self.maxpath()+1):
                print('x({0},{1})={2}'.format(i,j,self.xvar(i,j)))
        
        for i in range(1,self.N+1):
            for j in range(1, self.maxpath()+1):
                print('t({0},{1})={2}'.format(i,j,self.tvar(i,j)))
                
        for i in range(1,self.N+1):
            print('c({0})={1}'.format(i,self.cvar(i)))
        
        for i in range(1,self.N):
            for j in range(1, self.maxpath()+1):
                print('e({0},{1})={2}'.format(i,j,self.evar(i,j)))
        
        for i in range(1,self.N+1):
            for j in range(1, self.maxpath()+1):
                print('  -- -- --')
                for k in range(1, self.K+1):
                    print('f({0},{1},{2})={3}'.format(i,j,k,self.fvar(i,j,k)))
                    
        for i in range(1,self.N+1):
            for k in range(1, self.K+1):
                print('d0({0},{1})={2}'.format(i,k,self.d0var(i,k)))
                
        for i in range(1,self.N+1):
            for k in range(1, self.K+1):
                print('d1({0},{1})={2}'.format(i,k,self.d1var(i,k)))
    #
    #--------------------------------------------------------------------------
    def __init__(self, examples, featnames, satsolver, alg, cencd=0, npaths=2, pathBounds=0, recursive=0, quasipure=None, matchencd=0,trie = None, SKL=None, SKL_TOPO=None, tpldepth=3, ordertopologies=0, useassumps=500, verb=0, topv=0):
        super(PathsEnc,self).__init__()


       
        #print("#PathsEnc OPTIONS:")
        #print("# solver",satsolver)
        #print("# algorithm",alg)
        #print("# cardinality encoding",cencd)
        #print("# ub npaths", npaths)
        #print("# bounds paths", pathBounds)
        #print("# reiterative", recursive)
        #print("# quasipure", quasipure)
        #print("# match", matchencd)
        #print("# trie", trie)
        #print ("# sklearn topology:", end = '')
        #SKL_TOPO.print()
        #print()
        #print("# tpldepth", tpldepth)
        #print("# order topologies", ordertopologies)
        #print("# use assumps from # tplgs", useassumps)
        #print("# verb", verb)


        self.samples = examples
        self.fnames = featnames
        
        self.quasipure = quasipure
        
        self.ub  = -1
        self.model = []
        self.next_node = 1
        
        self.solver = satsolver
        self.algorithm = alg
        self.card = cencd
        self.recursive = recursive
        self.verbosity = verb
        self.M = len(examples)
        self.K = len(examples[0])-1
        self.N = npaths
        if self.N == -1:
            self.N = len(examples)
        self.top_npaths = self.N
        self.matchencd = matchencd
        self.trie = trie
        self.SKL = SKL
        self.SKL_TOPO = SKL_TOPO
        self.tpldepth = tpldepth
        self.ordertopologies = ordertopologies
        self.useassumps = useassumps
            
        self.pathBounds = pathBounds
        self.maxones = None
        self.maxzeros = None
        self.ubmxsat = None
        self.lbmxsat = None
        self.zerolbmxsat = None
        self.onelbmxsat = None
        
        self.create_variables(topv)
        #self.print_variables()
        #print('# top_id:', self.vpool._next())
    #
    #--------------------------------------------------------------------------
    # start the encoding in a cnf file
    def encode_paths(self, cnf):
        #
        cnf.append([-self.tvar(1,1)])
        cnf.append([-self.tvar(self.N,1)])
        
        for j in range(1, self.maxpath()+1):
            cnf.append([-self.xvar(1,j)])
        for j in range(1, self.maxpath()+1):
            cnf.append([self.xvar(self.N,j)])
        
        #print('#####\n right and left paths constraints\n', cnf.clauses)
        
        for i in range(1, self.N+1):
            for j in range(1, self.maxpath()):
                cnf.append([-self.tvar(i,j), self.tvar(i,j+1)])
        #print('#####\n if a path terminates then it is over until the end \n', cnf.clauses)
        
        for i in range(1, self.N):
            for j in range(1, self.maxpath()+1):
                if j > 1:
                    cnf.append([-self.evar(i,j), self.evar(i,j-1)])
                cnf.append([-self.evar(i,j), -self.xvar(i,j), self.xvar(i+1,j)])
                cnf.append([-self.evar(i,j), -self.xvar(i+1,j), self.xvar(i,j)])
                cnf.append([-self.evar(i,j), -self.tvar(i,j)])
                cnf.append([-self.evar(i,j), -self.tvar(i+1,j)])
        #print('#####\n contiguous paths must have equal part not yet ended\n', cnf.clauses)
        
        for i in range(1, self.N):
            # j=0
            cnf.append([self.evar(i,1), -self.xvar(i,1)])
            cnf.append([self.evar(i,1), self.xvar(i+1,1)])
            cnf.append([self.evar(i,1), -self.tvar(i,1)])
            cnf.append([self.evar(i,1), -self.tvar(i+1,1)])
            # j>0
            for j in range(1, self.maxpath()):
                cnf.append([-self.evar(i,j), self.evar(i,j+1), -self.xvar(i,j+1)])
                cnf.append([-self.evar(i,j), self.evar(i,j+1), self.xvar(i+1,j+1)])
                cnf.append([-self.evar(i,j), self.evar(i,j+1), -self.tvar(i,j+1)])
                cnf.append([-self.evar(i,j), self.evar(i,j+1), -self.tvar(i+1,j+1)])
        #print('#####\n contiguous paths must have a diferential part \n', cnf.clauses)
        
        for i in range(1, self.N):
            for j in range(1, self.maxpath()):
                cnf.append([self.evar(i,j), self.evar(i,j+1), self.xvar(i,j+1), self.tvar(i,j+1)])
                cnf.append([self.evar(i,j), self.evar(i,j+1), -self.xvar(i+1,j+1), self.tvar(i+1,j+1)])
        #print('#####\n contiguous paths end with disequalities \n', cnf.clauses)
        
        for i in range(1, self.N):
            cnf.append([-self.evar(i,self.maxpath())])
        #print('#####\n end of disaqualities \n', cnf.clauses)
    #
    #--------------------------------------------------------------------------
    #
    def encode_features(self, cnf):
    
        for i in range(1, self.N+1):
            for j in range(1, self.maxpath()+1):
                lits = [self.tvar(i,j)]
                for k in range(1, self.K+1):
                    lits.append(self.fvar(i,j,k))
                cnf.append(lits)
                cnf.extend(self.atMostOne(lits).clauses)
        #print('#####\n per position exactly 1 feature assigned or terminated \n', cnf.clauses)
        
        for i in range(1, self.N+1):
            for k in range(1, self.K+1):
                lits = []
                for j in range(1, self.maxpath()+1):
                    lits.append(self.fvar(i,j,k))
                cnf.extend(self.atMostOne(lits).clauses)
        #print('#####\n per path a ferature is assigned at most once \n', cnf.clauses)
        
        for i in range(1, self.N):
            for j in range(1, self.maxpath()+1):
                for k in range(1, self.K+1):
                    cnf.append([-self.evar(i,j), -self.fvar(i,j,k), self.fvar(i+1,j,k)])
                    cnf.append([-self.evar(i,j), self.fvar(i,j,k), -self.fvar(i+1,j,k)])
        #print('#####\n equal parts of contiguous paths must assign same feature \n', cnf.clauses)
        
        for i in range(1, self.N):
            # j=1
            for k in range(1, self.K+1):
                cnf.append([self.evar(i,1), -self.fvar(i,1,k), self.fvar(i+1,1,k)])
                cnf.append([self.evar(i,1), self.fvar(i,1,k), -self.fvar(i+1,1,k)])
            # j>1
            for j in range(2, self.maxpath()+1):
                for k in range(1, self.K+1):
                    cnf.append([-self.evar(i,j-1), self.evar(i,j), -self.fvar(i,j,k), self.fvar(i+1,j,k)])
                    cnf.append([-self.evar(i,j-1), self.evar(i,j), self.fvar(i,j,k), -self.fvar(i+1,j,k)])
        #print('#####\n diferential parts of contiguous paths must assign same feature \n', cnf.clauses)
    #
    #--------------------------------------------------------------------------
    #
    def encode_discriminates(self, cnf):
        
        for i in range(1, self.N+1):
            for k in range(1, self.K+1):
                lits = [-self.d0var(i,k)]
                for j in range(1, self.maxpath()+1):
                    aux = self.auxvar()
                    lits.append(aux)
                    cnf.append([aux, self.tvar(i,j), -self.fvar(i,j,k), -self.xvar(i,j)])
                    cnf.append([-aux, -self.tvar(i,j)])
                    cnf.append([-aux, self.fvar(i,j,k)])
                    cnf.append([-aux, self.xvar(i,j)])
                cnf.append(lits)
        #print('#####\n discriminate to 0 a feature in a path \n', cnf.clauses)
        
        for i in range(1, self.N+1):
            for k in range(1, self.K+1):
                lits = [-self.d1var(i,k)]
                for j in range(1, self.maxpath()+1):
                    aux = self.auxvar()
                    lits.append(aux)
                    cnf.append([aux, self.tvar(i,j), -self.fvar(i,j,k), self.xvar(i,j)])
                    cnf.append([-aux, -self.tvar(i,j)])
                    cnf.append([-aux, self.fvar(i,j,k)])
                    cnf.append([-aux, -self.xvar(i,j)])
                cnf.append(lits)
        #print('#####\n discriminate to 1 a feature in a path \n', cnf.clauses)
    #
    #--------------------------------------------------------------------------
    # encode the samples
    def encode_samples(self, cnf):
    
        if self.matchencd == 1:
            matchvar = lambda s,i: self.vpool.id('match{0} {1}'.format(s,i))
            for s in range(len(self.samples)):
                sample = self.samples[s]
                for i in range(1, self.N+1):
                    cl = [matchvar(sample,i)]
                    for k in range(1, self.K+1):
                        if sample[k-1] == 0:
                            cnf.append([-matchvar(s,i), -self.d0var(i,k)])
                            cl.append(self.d0var(i,k))
                        else:
                            cnf.append([-matchvar(s,i), -self.d1var(i,k)])
                            cl.append(self.d1var(i,k))
                    cnf.append(cl)
                    if sample[self.K] == 1:
                        cnf.append([-matchvar(sample,i), self.cvar(i)])
                    else:
                        cnf.append([-matchvar(sample,i),-self.cvar(i)])
            for i in range(1, self.N+1):
                cl = []
                for s in range(len(self.samples)):
                    cl.append(matchvar(s,i))
                cnf.append(cl)
            #print('#####\n samples (w/ match variables)\n', cnf.clauses,'\n\n')
            
        else:
        
            for sample in self.samples:
                for i in range(1, self.N+1):
                    cl = []
                    if sample[self.K] == 1:
                        cl.append(self.cvar(i))
                    else:
                        cl.append(-self.cvar(i))
                    for k in range(1, self.K+1):
                        if sample[k-1] == 0:
                            cl.append(self.d0var(i,k))
                        else:
                            cl.append(self.d1var(i,k))
                    cnf.append(cl)
            #print('#####\n samples \n', cnf.clauses,'\n\n')
    #
    #--------------------------------------------------------------------------
    # encode the quasipure
    def encode_quasipure(self, cnf):
        if self.quasipure is None:
            return
        
        for qpindex in self.quasipure.keys():
            k = qpindex+1
            fcount = self.quasipure[qpindex]
            for i in range(1, self.N+1):
                if fcount[0][0] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), self.xvar(i,self.maxpath()), self.cvar(i)])
                if fcount[0][1] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), self.xvar(i,j), -self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), self.xvar(i,self.maxpath()), -self.cvar(i)])
                if fcount[1][0] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), -self.xvar(i,self.maxpath()), self.cvar(i)])
                if fcount[1][1] == 0:
                    for j in range(1, self.maxpath()):
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), self.tvar(i,j+1)])
                        cnf.append([-self.fvar(i,j,k), -self.xvar(i,j), -self.cvar(i)])
                    cnf.append([-self.fvar(i,self.maxpath(),k), -self.xvar(i,self.maxpath()), -self.cvar(i)])
    #
    #--------------------------------------------------------------------------
    # encode the bounds of the paths
    def encode_path_ubmxsat(self, cnf):
        if self.ubmxsat == None:
            with RC2(WCNF(), trim=5, verbose=0) as rc2:
                #using variables of path 1 but could be any
                
                maxpath = min(self.top_npaths-1, self.K)
                
                if self.verbosity:
                    print('# Upper Bound MaxSAT maxpath testing:', maxpath)
                
                rc2.add_clause([-self.tvar(1, 1)])
                #print('WCNF:', [-self.tvar(1, 1)])
                
                for j in range(2, maxpath+1):
                    rc2.add_clause([-self.tvar(1,j)], weight=1)
                    #print('WCNF:', [-self.tvar(1,j)] ,'W1')
                    
                for j in range(2, maxpath):
                    rc2.add_clause([-self.tvar(1,j), self.tvar(1,j+1)])
                    #print('WCNF:', [-self.tvar(1,j), self.tvar(1,j+1)])
                    
                for j in range(1, maxpath+1):
                    lits = []
                    if j >1:
                        lits = [self.tvar(1,j)]
                    for k in range(1, self.K+1):
                        lits.append(self.fvar(1,j,k))
                    rc2.add_clause(lits)
                    #print('WCNF:', lits)
                    for cl in self.atMostOne(lits).clauses:
                        rc2.add_clause(cl)
                        #print('WCNF: am1 j:', j, cl)
                    
                for k in range(1, self.K+1):
                    lits = []
                    for j in range(1, maxpath+1):
                        lits.append(self.fvar(1,j,k))
                    for cl in self.atMostOne(lits).clauses:
                        rc2.add_clause(cl)
                        #print('WCNF: am1 k:', k, cl)
                
                matchvar = lambda s, j: self.vpool.id('match{0} {1}'.format(s,j))
                
                for s in range(1, len(self.samples)+1):
                    for j in range(1, maxpath+1):
                        #print('# matchvar({},{}) = {}'.format(s, j, matchvar(s,j)))
                        if j > 1:
                            rc2.add_clause([-matchvar(s,j), matchvar(s,j-1)])
                            #print('WCNF:', [-matchvar(s,j), matchvar(s,j-1)])
                        lits = [-matchvar(s,j)]
                        for k in range(1, self.K+1):
                            aux = self.auxvar()
                            lits.append(aux)
                            rc2.add_clause([-aux, self.fvar(1,j,k)])
                            #print('WCNF:', [-aux, self.fvar(1,j,k)])
                            if self.samples[s-1][k-1] == 0:
                                rc2.add_clause([-aux, -self.xvar(1,j)])
                                #print('WCNF:', [-aux, -self.xvar(1,j)])
                                rc2.add_clause([aux, -self.fvar(1,j,k), self.xvar(1,j)])
                                #print('WCNF:', [aux, -self.fvar(1,j,k), self.xvar(1,j)])
                            else:
                                rc2.add_clause([-aux, self.xvar(1,j)])
                                #print('WCNF:', [-aux, self.xvar(1,j)])
                                rc2.add_clause([aux, -self.fvar(1,j,k), -self.xvar(1,j)])
                                #print('WCNF:', [aux, -self.fvar(1,j,k), -self.xvar(1,j)])
                        rc2.add_clause(lits)
                        #print('WCNF:', lits)
                        
                matchvarother = lambda s, j: self.vpool.id('matchother{0} {1}'.format(s,j))
                            
                for s in range(1, len(self.samples)+1):
                    for j in range(1, maxpath+1):
                        #print('# matchvarother({},{}) = {}'.format(s, j, matchvarother(s,j)))
                        if j > 1:
                            rc2.add_clause([-matchvarother(s,j), matchvar(s,j-1)])
                            #print('WCNF:', [-matchvarother(s,j), matchvar(s,j-1)])
                        lits = [-matchvarother(s,j)]
                        for k in range(1, self.K+1):
                            aux = self.auxvar()
                            lits.append(aux)
                            rc2.add_clause([-aux, self.fvar(1,j,k)])
                            #print('WCNF:', [-aux, self.fvar(1,j,k)])
                            if self.samples[s-1][k-1] == 1:
                                rc2.add_clause([-aux, -self.xvar(1,j)])
                                #print('WCNF:', [-aux, -self.xvar(1,j)])
                                rc2.add_clause([aux, -self.fvar(1,j,k), self.xvar(1,j)])
                                #print('WCNF:', [aux, -self.fvar(1,j,k), self.xvar(1,j)])
                            else:
                                rc2.add_clause([-aux, self.xvar(1,j)])
                                #print('WCNF:', [-aux, self.xvar(1,j)])
                                rc2.add_clause([aux, -self.fvar(1,j,k), -self.xvar(1,j)])
                                #print('WCNF:', [aux, -self.fvar(1,j,k), -self.xvar(1,j)])
                        rc2.add_clause(lits)
                        #print('WCNF:', lits)
                
                matchposvar = lambda j: self.vpool.id('matchpos{0}'.format(j))
                matchnegvar = lambda j: self.vpool.id('matchneg{0}'.format(j))
                
                for j in range(1, maxpath+1):
                    litspos = [-matchposvar(j)]
                    litsneg = [-matchnegvar(j)]
                    litsother = [self.tvar(1, j)]
                    for s in range(1, len(self.samples)+1):
                        if self.samples[s-1][self.K] == 1:
                            litspos.append(matchvar(s,j))
                        else:
                            litsneg.append(matchvar(s,j))
                        litsother.append(matchvarother(s,j))
                    rc2.add_clause(litspos)
                    #print('WCNF:', litspos)
                    rc2.add_clause(litsneg)
                    #print('WCNF:', litsneg)
                    rc2.add_clause(litsother)
                    #print('WCNF:', litsother)
                    if j+1 <= maxpath:
                        rc2.add_clause([self.tvar(1, j+1), matchposvar(j)])
                        #print('WCNF:', [self.tvar(1, j+1), matchposvar(j)])
                        rc2.add_clause([self.tvar(1, j+1), matchnegvar(j)])
                        #print('WCNF:', [self.tvar(1, j+1), matchnegvar(j)])
                    rc2.add_clause([self.tvar(1, j), matchposvar(j), matchnegvar(j)])
                    #print('WCNF:', [self.tvar(1, j), matchposvar(j), matchnegvar(j)])
                    
                t = time.time()
                model = rc2.compute()
                if self.verbosity:
                    print('# Upper Bound MaxSAT time: {:.2f}'.format(time.time()-t))
                if model != None:
                    cost = rc2.cost
                    if self.verbosity:
                        #print('# Upper Bound MaxSAT cost:', cost)
                        print('# Upper Bound MaxSAT:', maxpath-cost)
                    #print('# Upper Bound MaxSAT model:',model)
                    self.ubmxsat = maxpath-cost
                else:
                    print('# No MaxSAT model found!')
                    self.ubmxsat = maxpath
                
        for i in range(1, self.N+1):
            for j in range(self.ubmxsat+1, self.maxpath()+1):
                cnf.append([self.tvar(i,j)])
                #print('# tvar set to 1 ubmxs {} {}'.format(i,j))
    #
    #--------------------------------------------------------------------------
    # encode the bounds of the paths
    def encode_pathBounds(self, cnf):
        maxzeros = 0
        maxones = 0
        if self.maxones == None or self.maxzeros == None:
            t = time.time()
            for sample in self.samples:
                nzeros = 0
                nones = 0
                for k in range(self.K):
                    if sample[k] == 0:
                        nzeros = nzeros + 1
                    else:
                        nones = nones + 1
                if nzeros > maxzeros:
                    maxzeros = nzeros
                if nones > maxones:
                    maxones = nones
            self.maxones = maxones
            self.maxzeros = maxzeros
            if self.verbosity > 0:
                print('# Upper Bound maxzeros:{} maxones:{}'.format(maxzeros,maxones))
                print('# Upper Bound time: {:.2f}'.format(time.time()-t))
        else:
            maxones = self.maxones
            maxzeros = self.maxzeros
        
                    
        for j in range(maxzeros+1, self.maxpath()+1):
            cnf.append([self.tvar(1,j)])
        for j in range(maxones +1, self.maxpath()+1):
            cnf.append([self.tvar(self.N,j)])
        
        
        #print('#####\n upper bound on the path 0 and 1\n', cnf.clauses)
        
        if self.pathBounds > 3:
            self.encode_path_ubmxsat(cnf)
        
        if self.pathBounds < 2:
            return
       
        if self.lbmxsat == None:
            with RC2(WCNF(), trim=5, verbose=0) as rc2:
                c = 1
                for f in range(1, self.K+1):
                    f0 = 2*f
                    f1 = f0+1
                    rc2.add_clause([-f0], weight=1)
                    rc2.add_clause([-f1], weight=1)
                    rc2.add_clause([-f0,-f1])
                    #print('# ms', [-f0, 'W1'], [-f1,'W1'], [-f0,-f1])
                for sample in self.samples:
                    cl = []
                    for f in range(self.K):
                        if sample[f] == 0:
                            cl.append(2*(f+1))
                        else:
                            cl.append(2*(f+1)+1)
                    if sample[self.K] == 1:
                        cl.append(c)
                    else:
                        cl.append(-c)
                    rc2.add_clause(cl)
                    #print('# ms', cl)
                  
                if self.pathBounds == 3:
                    lits0 = []
                    lits1 = []
                    for f in range(1, self.K+1):
                        lits0.append(2*f)
                        lits1.append(2*f+1)
                    rc2.add_clause(lits0)
                    rc2.add_clause(lits1)
                t = time.time()
                model = rc2.compute()
                if self.verbosity:
                    print('# Lower Bound MaxSAT time: {:.2f}'.format(time.time()-t))
                if model != None:
                    cost = rc2.cost
                    #print('# MaxSAT model:',model)
                    self.lbmxsat = cost
                else:
                    if self.verbosity:
                        print('# No MaxSAT model found!')
                    self.lbmxsat = 0
            if self.verbosity > 0:
                print('# Lower Bound MaxSAT: {0}'.format(self.lbmxsat))
        for i in range(2, self.N):
            for j in range(1, self.lbmxsat+1):
                cnf.append([-self.tvar(i,j)])
                #print('# tvar set to 0 lbmxs {} {}'.format(i,j))
                
        if self.zerolbmxsat == None:
            with RC2(WCNF(), trim=5, verbose=0) as rc2:
                c = self.K+1
                for f in range(1, self.K+1):
                    rc2.add_clause([-f], weight=1)
                    #print('# 0-ms', [-f, 'W1'])
                for sample in self.samples:
                    cl = []
                    for f in range(self.K):
                        if sample[f] == 1:
                            cl.append(f+1)
                   
                    if sample[self.K] == 1:
                        cl.append(c)
                    else:
                        cl.append(-c)
                    rc2.add_clause(cl)
                    #print('# 0-ms', cl)
                t =  time.time()
                model = rc2.compute()
                if self.verbosity:
                    print('# ZERO Lower Bound MaxSAT time: {:.2f}'.format(time.time()-t))
                if model != None:
                    cost = rc2.cost
                    #print('# ZERO MaxSAT model:',model)
                    self.zerolbmxsat = cost
                else:
                    print('# No ZERO MaxSAT model found!')
                    self.zerolbmxsat = 0
            if self.verbosity > 0:
                print('# ZERO Lower Bound MaxSAT: {0}'.format(self.zerolbmxsat))
        for j in range(2, self.zerolbmxsat+1):
            cnf.append([-self.tvar(1,j)])
            #print('# tvar set to 0 zero-lbmxs {} {}'.format(1,j))
        #i = 2
        #maxj = self.zerolbmxsat
        ##while self.lbmxsat < maxj:
        #while maxj > 0:
        #    for j in range(self.lbmxsat+1, maxj+1):
        #        cnf.append([-self.tvar(i,j)])
        #        #print('# tvar set to 0 zero-lbmxs {} {}'.format(i,j))
        #    for j in range(1, maxj):
        #        cnf.append([-self.xvar(i,j)])
        #        #print('# xvar set to 0 zero-lbmxs {} {}'.format(i,j))
        #        cnf.append([self.evar(i-1,j)])
        #        #print('# evar set to 1 zero-lbmxs {} {}'.format(i-1,j))
        #    i = i+1
        #    maxj = maxj-1
            
        if self.onelbmxsat == None:
            with RC2(WCNF(), trim=5, verbose=0) as rc2:
                c = self.K+1
                for f in range(1, self.K+1):
                    rc2.add_clause([-f], weight=1)
                    #print('# 1-ms', [-f, 'W1'])
                for sample in self.samples:
                    cl = []
                    for f in range(self.K):
                        if sample[f] == 0:
                            cl.append(f+1)
                    if sample[self.K] == 1:
                        cl.append(c)
                    else:
                        cl.append(-c)
                    rc2.add_clause(cl)
                    #print('# 1-ms', cl)
                t = time.time()
                model = rc2.compute()
                if self.verbosity:
                    print('# ONE Lower Bound MaxSAT time: {:.2f}'.format(time.time()-t))
                if model != None:
                    cost = rc2.cost
                    #print('# ZERO MaxSAT model:',model)
                    self.onelbmxsat = cost
                else:
                    print('# No ONE MaxSAT model found!')
                    self.onelbmxsat = 0
            if self.verbosity > 0:
                print('# ONE Lower Bound MaxSAT: {0}'.format(self.onelbmxsat))
        for j in range(2, self.onelbmxsat+1):
            cnf.append([-self.tvar(self.N,j)])
            #print('# tvar set to 0 one-lbmxs {} {}'.format(self.N,j))
        #i = self.N-1
        #maxj = self.onelbmxsat
        ##while self.lbmxsat < maxj:
        #while maxj > 0:
        #    for j in range(self.lbmxsat+1, maxj+1):
        #        cnf.append([-self.tvar(i,j)])
        #        #print('# tvar set to 0 one-lbmxs {} {}'.format(i,j))
        #    for j in range(1, maxj):
        #        cnf.append([self.xvar(i,j)])
        #        #print('# xvar set to 1 one-lbmxs {} {}'.format(i,j))
        #        cnf.append([self.evar(i,j)])
        #        #print('# evar set to 1 zero-lbmxs {} {}'.format(i,j))
        #    i = i-1
        #    maxj = maxj-1
    #
    #--------------------------------------------------------------------------
    # start the encoding in a cnf file
    def encode(self):
        cnf = CNF()
        self.encode_paths(cnf)
        self.encode_features(cnf)
        self.encode_discriminates(cnf)
        self.encode_samples(cnf)
        if self.quasipure is not None:
            self.encode_quasipure(cnf)
        if self.pathBounds > 0:
            self.encode_pathBounds(cnf)
        if self.verbosity:
            print('# n.clauses: {}'.format(len(cnf.clauses)))
        return cnf
    #
    #--------------------------------------------------------------------------
    def save_model(self, model, n):
        self.ub = n
        self.model = model
    #
    #--------------------------------------------------------------------------
    def sklearn_tree_to_sol(self, clf):
        n_nodes = clf.tree_.node_count
        children_left = clf.tree_.children_left
        children_right = clf.tree_.children_right
        feature = clf.tree_.feature
        threshold = clf.tree_.threshold
        value = clf.tree_.value
        rv = []
        for node_id in range(n_nodes):
            if (children_left[node_id] != children_right[node_id]):
                rv.append("a {} {}".format(self.fnames[feature[node_id]], node_id+1))
                rv.append("l {} {}".format(node_id+1, children_left[node_id]+1))
                rv.append("r {} {}".format(node_id+1, children_right[node_id]+1))
            else:
                rv.append("c {} {}".format(node_id+1, 0 if value[node_id][0].item(0) else 1))
        return rv
    #
    #--------------------------------------------------------------------------
    def report_solution_model(self,current,start,end,j):
        #print('Entering', current, start, end, j)
        r=1
        stop = False
        while not stop:
            if self.model[self.fvar(start,j,r)-1] > 0:
                print('a {} {}'.format(self.fnames[r-1],current))
                stop = True
            else:
                r = r+1
        
        left = self.next_node
        self.next_node = self.next_node+1
        print('l {} {}'.format(current, left))
        
        right = self.next_node
        self.next_node = self.next_node+1
        print('r {} {}'.format(current, right))
        
        divide = start+1
        stop = False
        while not stop:
            if self.model[self.xvar(divide,j)-1] > 0:
                stop = True
            else:
                divide = divide + 1
        #print('divide', divide)
        if j < self.maxpath():
            if self.model[self.tvar(start,j+1)-1] < 0:
                self.report_solution_model(left,start,divide-1,j+1)
            else:
                sgn = 0
                if self.model[self.cvar(start)-1] > 0:
                    sgn = 1
                print('c {} {}'.format(left,sgn))
                
            if self.model[self.tvar(divide,j+1)-1] < 0:
                self.report_solution_model(right,divide,end,j+1)
            else:
                sgn = 0
                if self.model[self.cvar(divide)-1] > 0:
                    sgn = 1
                print('c {} {}'.format(right,sgn))
        else:
            sgn = 0
            if self.model[self.cvar(start)-1] > 0:
                sgn = 1
            print('c {} {}'.format(left,sgn))
            sgn = 0
            if self.model[self.cvar(divide)-1] > 0:
                sgn = 1
            print('c {} {}'.format(right,sgn))
    #
    #--------------------------------------------------------------------------
    def report_solution(self, current = 1, next = 2):
            
        if self.ub == -1:
            print("UNSAT")
            return -1
        print('# minimum n paths:', self.ub, 'minimum n nodes:', 2*self.ub-1)
        #print('# model:', self.model)
        if self.model is None:
            SKL_DEC_TREE = self.sklearn_tree_to_sol(self.SKL)
            for l in SKL_DEC_TREE:
                print(l)
            return 1
        if self.model != []:
            self.N = self.ub
            self.next_node = next
            self.report_solution_model(current,1,self.ub,1)
        else:
            return None
        return 1
    #
    #--------------------------------------------------------------------------
    def tree_model(self,current,start,end,j):
        #print('Entering', current, start, end, j)
        node=[None,None,None]
        r=1
        stop = False
        while not stop:
            if self.model[self.fvar(start,j,r)-1] > 0:
                node[2] = self.fnames[r-1]
                #print('a {} {}'.format(self.fnames[r-1],current))
                stop = True
            else:
                r = r+1
        
        left = self.next_node
        self.next_node = self.next_node+1
        #print('l {} {}'.format(current, left))
        
        right = self.next_node
        self.next_node = self.next_node+1
        #print('r {} {}'.format(current, right))
        
        divide = start+1
        stop = False
        while not stop:
            if self.model[self.xvar(divide,j)-1] > 0:
                stop = True
            else:
                divide = divide + 1
        #print('divide', divide)
        if j < self.maxpath():
            if self.model[self.tvar(start,j+1)-1] < 0:
                node[0] = self.tree_model(left,start,divide-1,j+1)
            else:
                sgn = 0
                if self.model[self.cvar(start)-1] > 0:
                    sgn = 1
                #print('c {} {}'.format(left,sgn))
                node[0] = sgn
                
            if self.model[self.tvar(divide,j+1)-1] < 0:
                node[1]= self.tree_model(right,divide,end,j+1)
            else:
                sgn = 0
                if self.model[self.cvar(divide)-1] > 0:
                    sgn = 1
                #print('c {} {}'.format(right,sgn))
                node[1] = sgn
        else:
            sgn = 0
            if self.model[self.cvar(start)-1] > 0:
                sgn = 1
            #print('c {} {}'.format(left,sgn))
            node[0] = sgn
            sgn = 0
            if self.model[self.cvar(divide)-1] > 0:
                sgn = 1
            #print('c {} {}'.format(right,sgn))
            node[1] = sgn
        return node
    #
    #--------------------------------------------------------------------------
    def tree(self, current = 1, next = 1):
        if self.model != []:
            self.N = self.ub
            self.next_node = next
            return self.tree_model(current,1,self.ub,1)
        return -1
    #
    #--------------------------------------------------------------------------
    def satinit(self, cnf, useassumps = 0):
        if (self.recursive == 1 or useassumps ==1) and (self.solver == 'g3' or self.solver == 'g4'):
            return Solver(name = self.solver, bootstrap_with = cnf.clauses, incr = True)
        return Solver(name = self.solver, bootstrap_with = cnf.clauses)
    #
    #--------------------------------------------------------------------------
    def satsolve(self, oracle):
    
        if self.recursive == 0:
            return oracle.solve()
    
        if self.verbosity > 0:
            print('# Reiterative SAT Call Process:', end = '')
        
        maxpath = self.maxpath()
        #print('# maxpath in sat call:', maxpath)
        
        testsize = 2 #math.ceil(maxpath/2)
        #print('# start with paths of size:', testsize)
        
        indexes = []
        for i in range(1, self.N+1):
            indexes.append(testsize)
        
        while 1:
            #print('# indexes:', indexes)
        
            assumps = []
            refers = {}
            for i in range(1, self.N+1):
                if indexes[i-1] <= maxpath:
                    assumps.append(self.tvar(i, indexes[i-1]))
                    refers[self.tvar(i, indexes[i-1])] = i
            #print('# assumptions:', assumps)
            
            res = oracle.solve(assumptions=assumps)
            if not res:
                core = oracle.get_core()
                #print('# core:', core)
                #print('# -- UNSAT')
                if self.verbosity > 0:
                    print(' U', end = '')
                if core == None:
                    if self.verbosity > 0:
                        print('')
                    return res
                if len(core) < 0:
                    if self.verbosity > 0:
                        print('')
                    return res
                ncore = [-x for x in core]
                oracle.add_clause(ncore)
                for lit in core:
                    i = refers[lit]
                    if indexes[i-1] < maxpath+1:
                        indexes[i-1] = indexes[i-1] + 1
            else:
                #print('# -- SAT')
                #print('# model:', oracle.get_model())
                if self.verbosity > 0:
                    print(' S')
                return res
    #
    #--------------------------------------------------------------------------
    def solve_lssu(self):
        ubinit = self.N
        while self.N > 1 :
            if self.verbosity > 0:
                print('# Testing n paths:{} / n nodes:{}'.format(self.N,2*self.N-1))
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                self.N = self.N-2
            else:
                cnf = self.encode()
                if self.verbosity > 0:
                    print('# Testing n paths:{} / n nodes:{}'.format(self.N,2*self.N-1),'(n cls:{})'.format(len(cnf.clauses)))
                oracle = self.satinit(cnf)
                t = time.time()
                res = self.satsolve(oracle)
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    return
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    self.N = self.N-1
                oracle.delete()
    #
    #--------------------------------------------------------------------------
    def solve_lsus(self):
        stop = False
        ubinit = self.N
        self.N = 2
        while not stop :
            if self.verbosity > 0:
                print('# Testing n paths:{} / n nodes:{}'.format(self.N,2*self.N-1))
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                stop = True
            else:
                cnf = self.encode()
                if self.verbosity > 0:
                    print('# n cls:{}'.format(len(cnf.clauses)))
                oracle =self.satinit(cnf)
                t = time.time()
                res = self.satsolve(oracle)
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    self.N = self.N+1
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    stop = True
                oracle.delete()
    #
    #--------------------------------------------------------------------------
    def solve_bs(self):
        stop = False
        ubinit = self.N
        ub = self.N+1
        lb = 1
        while not stop :
            self.N = math.floor((ub + lb)/2)
            if self.verbosity > 0:
                print('# Testing n paths:{} / n nodes:{}'.format(self.N,2*self.N-1))
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
            else:
                cnf = self.encode()
                if self.verbosity > 0:
                    print('# n cls:{}'.format(len(cnf.clauses)))
                oracle =self.satinit(cnf)
                t = time.time()
                res = self.satsolve(oracle)
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
                oracle.delete()
            if ub == lb + 1:
                stop = True
    #
    #--------------------------------------------------------------------------
    def solve_prgsu(self):
        stop = False
        ubinit = self.N
        ub = self.N+1
        lb = 0
        p = 2
        i = 1
        while not stop :
            self.N = max(ub-pow(2,i), lb+1)
            if self.verbosity > 0:
                print('# Testing n paths:{} / n nodes:{}'.format(self.N,2*self.N-1))
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
                i = i+1
            else:
                cnf = self.encode()
                if self.verbosity > 0:
                    print('# n cls:{}'.format(len(cnf.clauses)))
                oracle =self.satinit(cnf)
                t = time.time()
                res = self.satsolve(oracle)
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                    i = 1
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
                    i = i+1
                oracle.delete()
            if ub == lb + 1:
                stop = True
    #
    #--------------------------------------------------------------------------
    def solve_prgus(self):
        stop = False
        ubinit = self.N
        ub = self.N+1
        lb = 1
        p = 2
        i = 1
        while not stop :
            self.N = min(lb+pow(2,i), ub-1)
            if self.verbosity > 0:
                print('# Testing n paths:{} / n nodes:{}'.format(self.N,2*self.N-1))
            if self.N == ubinit:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
                i = 1
            else:
                cnf = self.encode()
                if self.verbosity > 0:
                    print('# n cls:{}'.format(len(cnf.clauses)))
                oracle =self.satinit(cnf)
                t = time.time()
                res = self.satsolve(oracle)
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                    i = i+1
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
                    i = 1
                oracle.delete()
            if ub == lb + 1:
                stop = True
    #
    #--------------------------------------------------------------------------
    def settopology(self, root):
    
        assumps = []
        stack = [([],root)]
        
        #print()
        #root.print()
        #print()
        #print('# maxpath:', self.maxpath())
        
        npid = 1
        while len(stack) > 0:
            path, node = stack.pop()
            
            if node.card == 1:
                #print("{}:".format(npid), end = '')
                #for x in path:
                #    print(x, end = '')
                #print('t')
                for jindex in range(len(path)):
                    j = jindex+1
                    if j <= self.maxpath():
                        #print('x({0},{1})={2}:{3}'.format(npid,j,self.xvar(npid,j),path[jindex]))
                        if path[jindex] == 0:
                            assumps.append(-self.xvar(npid,j))
                        else:
                            assumps.append(self.xvar(npid,j))
                        #print('t({0},{1})={2}:{3}'.format(npid,j,self.tvar(npid,j), False))
                        assumps.append(-self.tvar(npid,j))
                if len(path) < self.maxpath():
                    for j in range(len(path)+1, self.maxpath()+1):
                        #print('t({0},{1})={2}:{3}'.format(npid,j,self.tvar(npid,j), True))
                        assumps.append(self.tvar(npid,j))
                npid = npid + 1
            else:
                if node.right is not None:
                    p = path[:]
                    p.append(1)
                    stack.append((p,node.right))
                if node.left is not None:
                    p = path[:]
                    p.append(0)
                    stack.append((p,node.left))
                if node.left is None and node.right is None:
                    for i in range((node.card+1)//2):
                        #print("{}:".format(npid), end = '')
                        #for x in path:
                        #    print(x, end = '')
                        #print('nt')
                        for jindex in range(len(path)):
                            j = jindex+1
                            if j < self.maxpath():
                                #print('x({0},{1})={2}:{3}'.format(npid,j,self.xvar(npid,j),path[jindex]))
                                if path[jindex] == 0:
                                    assumps.append(-self.xvar(npid,j))
                                else:
                                    assumps.append(self.xvar(npid,j))
                                #print('t({0},{1})={2}:{3}'.format(npid,j,self.tvar(npid,j), False))
                                assumps.append(-self.tvar(npid,j))
                        if len(path) < self.maxpath():
                            j = len(path)+1
                            #print('t({0},{1})={2}:{3}'.format(npid,j,self.tvar(npid,j), False))
                            assumps.append(-self.tvar(npid,j))
                        npid = npid + 1
        return assumps
    #
    #--------------------------------------------------------------------------
    def tplg_enum(self, tpldepth):
        return enum_topologies.enum(2*self.N-1, tpldepth)
    #
    #--------------------------------------------------------------------------
    def et_satsolve(self, cnf):
    
        useassumps = self.useassumps
        tpldepth = self.tpldepth
        oracle = None
        res = None
        
        if self.N <= 8:
            tpldepth = 2*self.N-1
        elif self.N <= 10:
            tpldepth = 3
          
        tgen = time.time()
        topologies = self.tplg_enum(tpldepth)
        #topologies = enum_topologies.enum(2*self.N-1, tpldepth)
        lt = len(topologies)
        if self.verbosity > 0:
            print('# n.topologies:',lt, 'tpldepth:', tpldepth)
        
        for tpl in topologies:
            enum_topologies.number_tree(tpl)
        
        scored = None
        if self.ordertopologies > 0:
            scored = [ (enum_topologies.diff_topologies(self.SKL_TOPO, tpl), i, tpl) for i, tpl in enumerate(topologies) ]
            scored.sort()
        else:
            scored = [ (1, i, tpl) for i, tpl in enumerate(topologies) ]
        
        if self.verbosity > 0:
            print('# Total topology generating and scoring time: {:.2f}'.format(time.time()-tgen))
        
        if lt > useassumps :
            oracle = self.satinit(cnf, True)
            print('# using assumption based solving')
        else:
            print('# no assumptions are used')
        
        satcalltimes = 0
        renewsatsolvertime = 0
        
        t = time.time()
        
        if self.verbosity > 0:
            print('# ', end = '')
        
        tobreak = False
        
        for i in range(len(scored)):
            if tobreak:
                break
            if self.verbosity > 0:
                print(';', end='')
            
            tpl = scored[i][2]
            assumps = self.settopology(tpl)
            
            if lt > useassumps:
                ti = time.time()
                res = oracle.solve(assumptions=assumps)
                if self.verbosity > 0:
                    tpi = time.time()-ti
                    satcalltimes += tpi
                    print('{:.2f}'.format(tpi), end='')
            else:
                ti = time.time()
                if oracle is not None:
                    oracle.delete()
                oracle = self.satinit(cnf, False)
                for aspt in assumps:
                    oracle.add_clause([aspt])
                tpi = time.time()-ti
                renewsatsolvertime += tpi
                
                ti = time.time()
                res = oracle.solve()
                if self.verbosity > 0:
                    tpi = time.time()-ti
                    satcalltimes += tpi
                    print('{:.2f}'.format(tpi), end='')
            
            if not res:
                if lt > useassumps:
                    core = oracle.get_core()
                    if core == None:
                        tobreak = True
                    elif len(core) < 0:
                        tobreak = True
                    else:
                        incore = False
                        for asp in assumps:
                            if asp in core:
                                incore = True
                                break
                        if not incore:
                            tobreak = True
            else:
                if self.verbosity > 0:
                    print('t{}'.format(i), end='')
                    #print('S', end='')
                tobreak = True
        if self.verbosity > 0:
            print('')
                
        if self.verbosity > 0:
            print('# Total topology test time: {:.2f}'.format(time.time()-t))
            print('# Total SAT Solving time {:.2f}'.format(satcalltimes))
            print('# Total Renew SAT Solver time {:.2f}'.format(renewsatsolvertime))
        
        return (oracle, res)
    #
    #--------------------------------------------------------------------------
    def solve_et_lssu(self):
        ubinit = self.N
        
        while self.N > 1 :
            if self.verbosity > 0:
                print('# Testing #paths =',self.N,'#nodes =', 2*self.N-1)
                
            if ubinit == self.N:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                self.N = self.N-1
            
            else:
                cnf = self.encode()
                
                (oracle, res) = self.et_satsolve(cnf)
                    
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    oracle.delete()
                    return 1
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    self.N = self.N-1
                    
                oracle.delete()
            
        return 1
    #
    #--------------------------------------------------------------------------
    def solve_et_lsus(self):
        stop = False
        ubinit = self.N
        self.N = 2
        while not stop :
            if self.verbosity > 0:
                print('# Testing #paths =',self.N)
                
            if self.N == ubinit:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                stop = True
            else:
                cnf = self.encode()
                    
                (oracle, res) = self.et_satsolve(cnf)
                        
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    self.N = self.N+1
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    stop = True
                    
                oracle.delete()
            
        return 1
    #
    #--------------------------------------------------------------------------
    def solve_et_bs(self):
        stop = False
        ubinit = self.N
        ub = self.N+1
        lb = 1
        while not stop :
            self.N = int((ub + lb)/2)
            if self.verbosity > 0:
                print('# Testing #paths =',self.N)
            
            if self.N == ubinit:
                if self.verbosity > 0:
                    print('# SAT (skl)')
                self.save_model(None, self.N)
                ub = self.N
            else:
                cnf = self.encode()
                
                (oracle, res) = self.et_satsolve(cnf)
                
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    lb = self.N
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.save_model(oracle.get_model(), self.N)
                    ub = self.N
                
                oracle.delete()
                
            if ub == lb + 1:
                stop = True
        
        return 1
    #
    #--------------------------------------------------------------------------
    def compute(self, depth = -1, spaces = ''):
        if self.algorithm == "lssu":
            self.solve_lssu()
        elif self.algorithm == "lsus":
            self.solve_lsus()
        elif self.algorithm == "bs":
            self.solve_bs()
        elif self.algorithm == "prgsu":
            self.solve_prgsu()
        elif self.algorithm == "prgus":
            self.solve_prgus()
        elif self.algorithm == "et-lssu":
            self.solve_et_lssu()
        elif self.algorithm == "et-lsus":
            self.solve_et_lsus()
        elif self.algorithm == "et-bs":
            self.solve_et_bs()
        else:
            print("Error:Unknown algorithm")
            exit(-1)
        return 1
    #
    #--------------------------------------------------------------------------
    def solve(self):
        if self.compute():
            return self.report_solution()
#
#==============================================================================
