#!/usr/bin/env python3
#-*- coding:utf-8 -*-
##
## main.py
##
##  Created on: Nov 21, 2019
##      Author: Antonio Morgado
##      E-mail: ajrmorgado@gmail.com
##
# Copyright (C) 2020, Antonio Morgado, Mikolas Janota
#
# This file is part of dtfinder.
#
# dtfinder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dtfinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dtfinder.  If not, see <https://www.gnu.org/licenses/>.
#
#==============================================================================
#
# required imports

import sys
import getopt
import os
import math
import time
import inspect
import functools
print = functools.partial(print, flush=True)

import dtfinderTop
import dtfinderTopDT1
import dtfinderTopPathsEncd

import mytrie


sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../scripts/'))

import enum_topologies
import sklearn.tree
#
#==============================================================================
def run_sk_learn(samples, verbosity):
    d = []
    v = []
    for s in samples:
        d.append( [ 1 if x==1 else -1 for x in s[:-1] ] )
        v.append(1 if s[-1]==1 else -1)
    clf = sklearn.tree.DecisionTreeClassifier(random_state=0)
    t0 = time.time()
    clf = clf.fit(d, v)
    depth =  clf.get_depth() + 1
    upper_bound =  2*clf.get_n_leaves() - 1
    if verbosity > 1:
        print ("# approx learn time: {:.2f} with upper bound {}".format(time.time() - t0, upper_bound))
    return depth, upper_bound, clf
#
#==============================================================================
def nnodes2npaths(nnodes):
    return (nnodes+1)//2
#
#==============================================================================
def npaths2nnodes(npaths):
    return 2*npaths-1
#
#==============================================================================
#
class SDTopPathsEnc(dtfinderTopPathsEncd.PathsEnc):
    #
    #--------------------------------------------------------------------------
    def __init__(self, examples, featnames, depth, satsolver, alg, cencd=0, depthorcls=0, npaths=2, pathBounds=0, recursive=0, quasipure=None, matchencd=0,trie = None, SKL=None, SKL_TOPO=None, tpldepth=3, tplenum=0, ordertopologies=0, useassumps=500, verb=0, topv=0):
    
        #if verb >= 0 :
        #    print("#SDTopPathsEnc OPTIONS:")
        #    print("# ub depth", depth)
        #    print("# solver",satsolver)
        #    print("# algorithm",alg)
        #    print("# cardinality encoding",cencd)
        #    print("# ub npaths", npaths)
        #    print("# bounds paths", pathBounds)
        #    print("# reiterative", recursive)
        #    print("# quasipure", quasipure)
        #    print("# match", matchencd)
        #    print("# trie", trie)
        #    print ("# sklearn topology:", end = '')
        #    SKL_TOPO.print()
        #    print()
        #    print("# tpldepth", tpldepth)
        #    print("# tplenum", tplenum)
        #    print("# order topologies", ordertopologies)
        #    print("# use assumps from # tplgs", useassumps)
        #    print("# verb", verb)
    
        super(SDTopPathsEnc,self).__init__(examples, featnames, satsolver, alg, cencd, npaths, pathBounds, recursive, quasipure, matchencd, trie, SKL, SKL_TOPO, tpldepth, ordertopologies, useassumps, verb, topv)
        
        #examples, featnames, satsolver, alg, cencd=0, npaths=2, pathBounds=0, recursive=0, quasipure=None, matchencd=0,trie = None, SKL=None, SKL_TOPO=None, tpldepth=3, ordertopologies=0, useassumps=500, verb=0, topv=0
        
        self.depthorcls = depthorcls
        self.ubdepth = depth
        self.ubpaths = npaths
        self.depth = depth
        self.tplenum = tplenum
    #
    #--------------------------------------------------------------------------
    def encode_maxpathsize(self, cnf):
        if self.depth > self.maxpath():
            #print('# depth:', self.depth, 'is bigger than maxpath()', self.maxpath())
            return
        for i in range(1,self.N+1):
            cnf.append([self.tvar(i,self.depth)])
            #print('# path', i,'maxpathsize cl:', [self.tvar(i,self.depth)])
        if self.depthorcls == 1:
            cl = []
            for i in range(1,self.N+1):
                cl.append(-self.tvar(i,self.depth-1))
            print('# depth-or-cl', cl)
            cnf.append(cl)
    #
    #--------------------------------------------------------------------------
    def encode(self):
        cnf = super(SDTopPathsEnc,self).encode()
        self.encode_maxpathsize(cnf)
        if self.verbosity > 0:
            print('# n cls:{}'.format(len(cnf.clauses)))
        return cnf
    #
    #--------------------------------------------------------------------------
    def setN(self, nnodes=0, npaths=0):
        if nnodes == 0:
            if npaths != 0:
                self.N = npaths
                return
        if npaths == 0:
            if nnodes != 0:
                self.N = nnodes2npaths(nnodes)
                return
        print("ERROR:SDTopPathsEnc.setN(X) does not accept both nnodes and npaths to be 0.")
        exit(-1)
    #
    #--------------------------------------------------------------------------
    def getNPaths(self):
        return self.N
    #
    #--------------------------------------------------------------------------
    def getNNodes(self):
        return npaths2nnodes(self.N)
    #
    #--------------------------------------------------------------------------
    def tplg_enum(self, tpldepth):
    
        print("# Calling the enumeration of tolpologies in SDTopPathsEnc")
        
        if self.tplenum == 1:
            return enum_topologies.enum2_eq_depth(2*self.N-1, self.depth, tpldepth)
        return enum_topologies.enum(2*self.N-1, tpldepth)
#
#==============================================================================
#
class SDTopSolver(object):
    #
    #--------------------------------------------------------------------------
    def __init__(self, examples, featnames, depth, encoding, satsolver, alg, cencd=0, depthorcls=0, nnodes=3, npaths=2, pathBounds=0, recursive=0, quasipure=None, matchencd=0,trie = None, SKL=None, SKL_TOPO=None, tpldepth=3, tplenum=0, ordertopologies=0, useassumps=500, verb=0, topv=0):
        
        #if verb >= 0 :
        #    print("#SDTopSolver OPTIONS:")
        #    print("# ub depth", depth)
        #    print("# encoding", encoding)
        #    print("# solver",satsolver)
        #    print("# algorithm",alg)
        #    print("# cardinality encoding",cencd)
        #    print("# ub nnodes", nnodes)
        #    print("# ub npaths", npaths)
        #    print("# bounds paths", pathBounds)
        #    print("# reiterative", recursive)
        #    print("# quasipure", quasipure)
        #    print("# match", matchencd)
        #    print("# trie", trie)
        #    print ("# sklearn topology:", end = '')
        #    SKL_TOPO.print()
        #    print()
        #    print("# tpldepth", tpldepth)
        #    print("# tplenum", tplenum)
        #    print("# order topologies", ordertopologies)
        #    print("# use assumps from # tplgs", useassumps)
        #    print("# verb", verb)
    
    
        self.encoding = encoding
        self.verbosity = verb
        self.algorithm = alg
        
        if encoding == 0:
            print("SDTopSolver.__init__::Encoding yet to be implemented.")
            exit(-1)
        elif encoding == 1:
            if nnodes != 3:
                npaths = nnodes2npaths(nnodes)
            self.encoder = SDTopPathsEnc(examples, featnames, depth, satsolver, alg, cencd, depthorcls, npaths, pathBounds, recursive, quasipure, matchencd, trie, SKL, SKL_TOPO, tpldepth, tplenum, ordertopologies, useassumps, verb, topv)
        else:
            print("Error: encoding not invented.")
            exit(-1)
        
        
    #
    #--------------------------------------------------------------------------
    def solve_bu_lsus(self):
        stop = False
        self.encoder.depth = 2
        self.encoder.setN(npaths=self.encoder.depth)
        while not stop :
            if self.verbosity > 0:
                print('# Testing (depth:{}) n paths:{} / n nodes:{}'.format(self.encoder.depth, self.encoder.getNPaths(), self.encoder.getNNodes()))
            if self.encoder.depth == self.encoder.ubdepth and self.encoder.getNPaths() == self.encoder.ubpaths:
                if self.verbosity > 0:
                    print('# SAT (SKL)')
                self.encoder.save_model(None, self.encoder.getNPaths())
                stop = True
            else:
                cnf = self.encoder.encode()
                if self.verbosity > 0:
                    print('# n cls:{}'.format(len(cnf.clauses)))
                if self.encoding == 1:
                    if self.encoder.pathBounds > 1:
                        docontinue = False
                        if self.encoder.lbmxsat is not None:
                            if self.encoder.depth < self.encoder.lbmxsat:
                                self.encoder.depth = self.encoder.lbmxsat
                                docontinue = True
                        if self.encoder.zerolbmxsat is not None:
                            if self.encoder.depth < self.encoder.zerolbmxsat:
                                self.encoder.depth = self.encoder.zerolbmxsat
                                docontinue = True
                        if self.encoder.onelbmxsat is not None:
                            if self.encoder.depth < self.encoder.onelbmxsat:
                                self.encoder.depth = self.encoder.onelbmxsat
                                docontinue = True
                        if docontinue:
                            self.encoder.setN(npaths = self.encoder.depth)
                            print('# UNSAT')
                            continue
                oracle = self.encoder.satinit(cnf)
                t = time.time()
                res = self.encoder.satsolve(oracle)
                if self.verbosity > 0:
                    print('# SAT call time: {:.2f}'.format(time.time()-t))
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    self.encoder.setN(npaths = self.encoder.getNPaths()+1)
                    if self.encoder.getNPaths() > pow( 2, self.encoder.depth-1):
                        self.encoder.depth = self.encoder.depth + 1
                        self.encoder.setN(npaths = self.encoder.depth)
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.encoder.save_model(oracle.get_model(), self.encoder.getNPaths())
                    stop = True
                oracle.delete()
    #
    #--------------------------------------------------------------------------
    def solve_et_bu_lsus(self):
        stop = False
        self.encoder.depth = 2
        self.encoder.setN(npaths=self.encoder.depth)
        while not stop :
            if self.verbosity > 0:
                print('# Testing (depth:{}) n paths:{} / n nodes:{}'.format(self.encoder.depth, self.encoder.getNPaths(), self.encoder.getNNodes()))
            if self.encoder.depth == self.encoder.ubdepth and self.encoder.getNPaths() == self.encoder.ubpaths:
                if self.verbosity > 0:
                    print('# SAT (SKL)')
                self.encoder.save_model(None, self.encoder.getNPaths())
                stop = True
            else:
                cnf = self.encoder.encode()
                if self.encoding == 1:
                    if self.encoder.pathBounds > 1:
                        docontinue = False
                        if self.encoder.lbmxsat is not None:
                            if self.encoder.depth < self.encoder.lbmxsat:
                                self.encoder.depth = self.encoder.lbmxsat
                                docontinue = True
                        if self.encoder.zerolbmxsat is not None:
                            if self.encoder.depth < self.encoder.zerolbmxsat:
                                self.encoder.depth = self.encoder.zerolbmxsat
                                docontinue = True
                        if self.encoder.onelbmxsat is not None:
                            if self.encoder.depth < self.encoder.onelbmxsat:
                                self.encoder.depth = self.encoder.onelbmxsat
                                docontinue = True
                        if docontinue:
                            self.encoder.setN(npaths = self.encoder.depth)
                            print('# UNSAT')
                            continue
                    
                (oracle, res) = self.encoder.et_satsolve(cnf)
                        
                if not res:
                    if self.verbosity > 0:
                        print('# UNSAT')
                    self.encoder.setN(npaths = self.encoder.getNPaths()+1)
                    if self.encoder.getNPaths() > pow( 2, self.encoder.depth-1):
                        self.encoder.depth = self.encoder.depth + 1
                        self.encoder.setN(npaths = self.encoder.depth)
                else:
                    if self.verbosity > 0:
                        print('# SAT')
                    self.encoder.save_model(oracle.get_model(), self.encoder.getNPaths())
                    stop = True
                oracle.delete()
                    
        return 1
    #
    #--------------------------------------------------------------------------
    def compute(self):
        if self.algorithm == "bu-lsus":
            self.solve_bu_lsus()
        elif self.algorithm == "et-bu-lsus":
            self.solve_et_bu_lsus()
        else:
            print("SDTopSolver.compute:: Error - Unknown algorithm")
            return -1
        return 1
    #
    #--------------------------------------------------------------------------
    def solve(self):
        if self.compute():
            print('# minimum depth:', self.encoder.depth)
            return self.encoder.report_solution()
#
#==============================================================================
#
def main(argv):
    
    algorithm = 'et-bu-lsus'
    cencd = 0
    depth = -1
    dintriesetsamples = 1
    depthorcls = 0
    encoding = 1
    inputfile = 's4.txt'
    lbDT1encd = 0
    matchencd = 0
    nnodes = -1
    npaths = -1
    ordertopologies = 0
    pathbounds = 0
    quasipure = None
    tpldepth = 3
    tplenum = 0
    reiterative = 0
    solver = 'g3'
    ubDT1encd = 0
    useassumps = 500
    verbosity = 0
    
    try:
        opts, args = getopt.getopt(argv,"a:b:c:d:e:hi:lm:o:p:qt:rs:uvx:",["algorithm", "cardenc", "depthorcls", "encoding","help", "ifile", "lbDT1", "match", "ordertop", "pathbounds", "quasipure", "tpldepth", "tplenum", "reiterative", "solver", "ubDT1", "useasptpl", "verbose"])
    except getopt.GetoptError:
        print('Error in input parameters, for help pease use: $dtfinder.py -h')
        sys.exit(2)
        
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print('sddtfinder.py (-i|-ifile) <inputfile> [<options>]')
            print(' (-a|--algorithm) <str>')
            print('    sets the main search algorithm: ')
            #print('      td-bs ---- (default) top-down for depth with binary search for #nodes')
            #print('      td-lsus -- top-down for depth with linear search unsat-sat for #nodes')
            #print('      bu-bs ---- bottom-up for depth with binary search for #nodes')
            print('      bu-lsus    -- bottom-up for depth with linear search unsat-sat for #nodes')
            print('      et-bu-lsus -- same as bu-lsus but using topologies')
            print(' (-b|--pathbounds) <int>')
            print('    sets the type of path size bounds to use in Paths encoding')
            print('      0 -- no bounds are used')
            print('      1 -- upper bounds of maxones and maxzeros')
            print('      2 -- same as 1 plus lower bounds using maxsat')
            print('      3 -- same as 2 plus using or constraints in maxsat')
            print('      4 -- same as 3 plus upper bounds using maxsat (not yet done)')
            print(' (-c|--cardenc) <int>')
            print('    sets the cardinality encoding (as allowed by PySAT):')
            print('      0 ---- Pairwise(default)')
            print('      1 ---- Seq Counter')
            print('      ... -- (any allowed by pysat)')
            print('      8 ---- kMTot')
            print(' (-d|--depthorcls) <int>')
            print('     sets the or clauses making atleast one path with max depth')
            print('      0  --  none (default)')
            print('      1  --  add 1 or clause for all paths')
            print(' (-e|--encoding) <int>')
            print('    sets the encoding:')
            print('      0 -- DT1')
            print('      1 -- (default) Paths encoding')
            print(' (-h|--help)')
            print('    this help message')
            print(' (-l|--lbDT1)')
            print('    encode the lower bound of the DT1 encoding (section 3.3)')
            print(' (-m|--match) <int>')
            print('    sets the encoding of discrimination in the Paths encoding')
            print('      0 -- no match variables are created')
            print('      1 -- creates a match variable for each path and each sample')
            print('      2 -- creates a match variable for each path and each sample using a trie (DT1 only)')
            print('      3 -- creates a match variable for each path and each sample using a compressed trie (DT1 only)')
            print(' (-o|--ordertop) <int>')
            print('    order de topologies according to similarity to sklearn topology')
            print('      0 -- no ordering ')
            print('      1 -- use ordering')
            print(' (-p|--tplenum) <int>')
            print('    type of topology enumeration used')
            print('      0 -- (default) complete enumeration, that is all (partial) topologies')
            print('      1 -- enumerates (partial) topologies considering depth')
            print(' (-q|--quasipure)')
            print('    sets quasipure constraints in the Paths encoding')
            print('      0 -- no quasipure constraints')
            print('      1 -- use quasipure constraints')
            print(' (-t|--tpldepth) <int>')
            print('    maximumn depth of partial topologies enumerated that have more than 21 nodes')
            print(' (-r|--reiterative)')
            print('    ...(to be done)')
            print(' (-s|--solver) <str>')
            print('    the sat solver to be used, g3(default), g4, cd, lgl, mcb, ... (any allowed by pysat)')
            print(' (-u|--ubDT1)')
            print('    encode the upper bound of the DT1 encoding (section 3.3)')
            print(' (-v|--verbose')
            print('    set the verbose mode')
            print(' (-x|--useasptpl) <int>')
            print('    minimum number of (partial) topologies enumerated to start using assumptions')
            sys.exit()
            
        elif opt in ("-a", "--algorithm"):
            algorithm = arg
        elif opt in ("-b", "--pathbounds"):
            pathbounds = int(arg)
        elif opt in ("-c", "--cardenc"):
            cencd = int(arg)
        elif opt in ("-d", "--depthorcls"):
            depthorcls = int(arg)
        elif opt in ("-e", "--encoding"):
            encoding = int(arg)
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt in ("-l", "--lbDT1"):
            lbDT1encd = 1
        elif opt in ("-m", "--match"):
            matchencd = int(arg)
        elif opt in ("-o", "--ordertop"):
            ordertop = int(arg)
        elif opt in ("-p", "--tplenum"):
            tplenum = int(arg)
        elif opt in ("-q", "--quasipure"):
            quasipure = {}
        elif opt in ("-t", "--tlpdepth"):
            tlpdepth = int(arg)
        elif opt in ("-r", "--reiterative"):
            reiterative = 1
        elif opt in ("-s", "--solver"):
            solver = arg
        elif opt in ("-u", "--ubDT1"):
            ubDT1encd = 1
        elif opt in ("-v", "--verbose"):
            verbosity = 1
        elif opt in ("-x", "--useasptpl"):
            useassumps = int(arg)
    

    examples, featNames = dtfinderTop.read_examples_file(inputfile, verbosity)
    
    if verbosity > 0:
        print('# nfeat:{} nsamples:{}'.format(len(examples[0])-1, len(examples)))
       
    examples, featNames, quasipure = dtfinderTop.preprocess(examples, featNames, quasipure, verbosity)
        
    
    trie = None
    if matchencd > 1:
        t0 = time.time()
        trie = mytrie.Trie()
        if dintriesetsamples > 0:
            trie.setsamplesdin(examples)
        else:
            trie.setsamples(examples)
        if verbosity > 0:
            print('# Trie building time: {:.2f}'.format(time.time()-t0))
            print('# n.nodes in trie: {} (compression rate: {:.2f})'.format(trie.nid, (trie.nid / ((len(examples[0])-1)* len(examples)) ) ))
          
    depth, nnodes, SKL = run_sk_learn(examples, verbosity)
    npaths = (nnodes+1)//2
           
    SKL_TOPO = enum_topologies.sklearn2topology(SKL)
    enum_topologies.number_tree(SKL_TOPO)
           
    if verbosity > 0:
        print('# ub-depth:', depth, ' ub-paths:', npaths, 'ub-nodes:', nnodes)
        print ("# sklearn topology:", end = '')
        SKL_TOPO.print()
        print()
        
    if verbosity > 1 :
        print("#OPTIONS:")
        print("# ub depth", depth)
        print("# encoding", encoding)
        print("# solver",solver)
        print("# algorithm",algorithm)
        print("# cardinality encoding",cencd)
        print("# ub nnodes", nnodes)
        print("# ub npaths", npaths)
        print("# bounds paths", pathbounds)
        print("# reiterative", reiterative)
        print("# quasipure", quasipure)
        print("# match", matchencd)
        print("# trie", trie)
        print ("# sklearn topology:", end = '')
        SKL_TOPO.print()
        print()
        print("# tpldepth", tpldepth)
        print("# tplenum", tplenum)
        print("# order topologies", ordertopologies)
        print("# use assumps from # tplgs", useassumps)
        print("# verb", verbosity)
    
    solver = SDTopSolver(examples, featNames, depth, encoding, solver, algorithm, cencd, depthorcls, nnodes, npaths, pathbounds, reiterative, quasipure, matchencd, trie, SKL, SKL_TOPO, tpldepth, tplenum, ordertopologies, useassumps, verbosity)
    solver.solve()
        


#
#==============================================================================

if __name__ == "__main__":
    main(sys.argv[1:])
