#!/usr/bin/env python3
# File:  enum.py
# Author:  mikolas
# Created on:  Wed Feb 12 12:07:00 WET 2020
# Copyright (C) 2020, Mikolas Janota
# Copyright (C) 2020, Antonio Morgado, Mikolas Janota
#
# This file is part of dtfinder.
#
# dtfinder is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dtfinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dtfinder.  If not, see <https://www.gnu.org/licenses/>.

import sys
import queue
import math
import os,inspect

sys.path.insert(0, os.path.join(os.path.realpath(os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])), '../greedy/'))
import run_sklearn

# memoization is not used because we need to give IDs to each subtree
# so sharing doesn't work

class Node:
    def __init__(self, card, left, right):
        self.card = card
        self.left = left
        self.right = right

    def to_string(self):
        rv = ''
        if self.node_id is not None:
            rv += '{}:'.format(self.node_id)
        if self.card > 1 and self.left is None:
            return rv + '(#{})'.format(self.card)
        if self.card == 1:
            return rv + '()'
        else:
            return rv + '({}{})'.format(self.left.to_string(), self.right.to_string())

    def print(self):
        if self.node_id is not None:
            print('{}:'.format(self.node_id), end='')
        if self.card > 1 and self.left is None:
            print('(#{})'.format(self.card), end='')
            return
        print('(', end='')
        if self.card > 1:
            self.left.print()
            self.right.print()
        print(')', end='')

def get_card(t):
    if t is None:
        return 0
    if t.card is not None:
        return t.card
    else:
        return 1 + get_card(t.left) + get_card(t.right)

def show_nfolhas(root):
    q = queue.Queue()
    q.put(root)
    count = 1
    sum_leafs = 0
    while not q.empty():
        n = q.get()
        if n.left is None:
            sum_leafs += (n.card+1) // 2
        else:
            q.put(n.left)
            q.put(n.right)
        print("{} leafs {}".format(n.node_id, sum_leafs))

# define the difference between two to topologies
# if topologies are exactly the same,  the return value is zero, otherwise it's a positive number
def diff_topologies(t1,t2):
    factor = 0.75
    def diff_topologies_rec(t1,t2,decay):
        if t1 is None:
            return get_card(t2)*decay
        if t2 is None:
            return get_card(t1)*decay
        # if (t1.card > 1 and t1.left is None) or (t2.card > 1 and t2.left is None):
        #     return abs(get_card(t1) - get_card(t2))
        return diff_topologies_rec(t1.left, t2.left,decay*factor) + diff_topologies_rec(t1.right, t2.right,decay*factor) 
    return diff_topologies_rec(t1,t2,1)

def sklearn2topology(clf):
    children_left = clf.tree_.children_left
    children_right = clf.tree_.children_right

    def sklearn2topology_rec(node_id, memz):
        if node_id in memz:
            return memz[node_id]

        if (children_left[node_id] != children_right[node_id]):
            lt = sklearn2topology_rec(children_left[node_id], memz)
            rt = sklearn2topology_rec(children_right[node_id], memz)
            rv = Node(1+lt.card+rt.card, lt, rt)
        else:
            rv = Node(1, None, None)
        memz[node_id] = rv
        return rv

    return sklearn2topology_rec(0, dict())


def enum1_mem(card, d, m):
    # if (card, d) in m:
    #     return m[(card, d)]

    if card == 1 or d <= 0:
        rv = [ Node(card, None, None) ]
    else:
        if d == 1:
            if card == 3:
                rv =  [ Node(card, Node(1, None, None), Node(1, None, None)) ]
            elif card == 5:
                rv =  [ Node(card, Node(1, None, None), Node(3, None, None)),\
                         Node(card, Node(3, None, None), Node(1, None, None)) \
                         ]
            else: rv =  [\
                    Node(card,Node(1, None, None), Node(999, None, None)),\
                     Node(card,Node(999, None, None), Node(999, None, None)),\
                     Node(card,Node(999, None, None), Node(1, None, None))\
                     ]
        else:
            rv = []
            for left_card in range(1, card, 2):
                left_tps = enum1_mem(left_card, d-1, m)
                right_tps = enum1_mem(card - left_card - 1, d-1, m)
                for lt in left_tps:
                    for rt in right_tps:
                        rv.append(Node(card, lt, rt))

    # m[(card,d)] = rv
    return rv

def enum1(card, d):
    return enum1_mem(card, d, dict())

def enum_mem(card, d, m):
    # if (card, d) in m:
    #     return m[(card, d)]

    if card == 1 or d <= 0:
        rv = [ Node(card, None, None) ]
        m[(card,d)] = rv
        return rv

    rv = []
    for left_card in range(1, card, 2):
        left_tps = enum_mem(left_card, d-1, m)
        right_tps = enum_mem(card - left_card - 1, d-1, m)
        for lt in left_tps:
            for rt in right_tps:
                rv.append(Node(card, lt, rt))

    # m[(card,d)] = rv
    return rv

def enum(card, d):
    return enum_mem(card, d, dict())

# add numbers to each node by BFS
def number_tree(root):
    q = queue.Queue()
    q.put(root)
    count = 1
    saw_unfinished = False
    while not q.empty():
        n = q.get()
        if n is None:
            saw_unfinished = True
            continue
        assert(not saw_unfinished) # check that no Node is given an id after an unexpanded sub-tree
        n.node_id = count
        count += 1
        if n.card > 1: # potentially insert None, only for the purpose of the assertion above
            q.put(n.left)
            q.put(n.right)

def my_log2(i):
    rv = 0
    while i > 0:
        i = i >> 1
        rv += 1
    return rv

def depth_lb(t):
    if t is None:
        return 0
    if t.left is None and t.card > 1:
        return math.ceil(math.log2(t.card+1))
    return 1 + max(depth_lb(t.left), depth_lb(t.right))


def depth_ub(t):
    if t is None:
        return 0
    if t.left is None and t.card > 1:
        return (t.card + 1)/2
    return 1 + max(depth_ub(t.left), depth_ub(t.right))

# max_depth = 1 implies a leaf
# for now we do not have is a cutoff depth for enumeration
def enum1_depth(card, depth):
    if 2**depth - 1 < card:
        err("max_depth can't fit the nodes required {} {}".format(2**depth - 1, card))
    d = card
    ls = enum1(card, d)
    return [ t for t in ls if depth_lb(t) == depth ]

def enum2_leq_depth(card, depth, cutoff):
    if 2**depth - 1 < card:
        # depth not sufficient to accommodate card nodes
        return []
    if card == 1 or depth == 1:
        # this is a leaf (actual leaf in the topology)
        return [ Node(1, None, None) ]
    if cutoff == 0:
        # this set of trees is nonempty due to the test on the first line
        return [ Node(card, None, None) ]
    rv = []
    for left_card in range(1, card, 2):
        left_tps = enum2_leq_depth(left_card, depth - 1, cutoff - 1)
        right_tps = enum2_leq_depth(card - left_card - 1, depth - 1, cutoff - 1) if left_tps else [] # avoid calculating right_tps if left_tps already empty
        for lt in left_tps:
            for rt in right_tps:
                rv.append(Node(card, lt, rt))
    return rv


def enum2_eq_depth(card, depth, cutoff):
    if 2**depth - 1 < card or card < 2*depth - 1:
        # depth not sufficient to accommodate card nodes
        # or there's no way to achive this depth
        return []
    if card == 1:
        # this is a leaf (actual leaf in the topology)
        return [ Node(1, None, None) ]
    if cutoff == 0:
        # this set of trees is nonempty due to the test on the first line
        return [ Node(card, None, None) ]
    rv = []
    for left_card in range(1, card, 2):
        # 1) partion left topology depth = depth - 1, right topology depth < depth - 1
        left_tps = enum2_eq_depth(left_card, depth - 1, cutoff - 1)
        right_tps = enum2_leq_depth(card - left_card - 1, depth - 2, cutoff - 1) if left_tps else [] # avoid calculating right_tps if left_tps already empty
        for lt in left_tps:
            for rt in right_tps:
                rv.append(Node(card, lt, rt))
        # 2) partion left topology depth < depth - 1, right topology depth = depth - 1
        left_tps = enum2_leq_depth(left_card, depth - 2, cutoff - 1)
        right_tps = enum2_eq_depth(card - left_card - 1, depth - 1, cutoff - 1) if left_tps else [] # avoid calculating right_tps if left_tps already empty
        for lt in left_tps:
            for rt in right_tps:
                rv.append(Node(card, lt, rt))
        # 3) partion left topology depth = depth - 1, right topology depth = depth - 1
        left_tps = enum2_eq_depth(left_card, depth - 1, cutoff - 1)
        right_tps = enum2_eq_depth(card - left_card - 1, depth - 1, cutoff - 1) if left_tps else [] # avoid calculating right_tps if left_tps already empty
        for lt in left_tps:
            for rt in right_tps:
                rv.append(Node(card, lt, rt))

    # left_tps_1 = enum2_depth(1, depth - 1, cutoff - 1)
    # right_tps_1 = enum2_depth(card - 1 - 1, depth - 1, cutoff - 1)

    # left_tps_3 = enum2_depth(3, depth - 1, cutoff - 1)
    # right_tps_3 = enum2_depth(card - 3 - 1, depth - 1, cutoff - 1)

    # left_tps_5 = enum2_depth(5, depth - 1, cutoff - 1)
    # right_tps_5 = enum2_depth(card - 5 - 1, depth - 1, cutoff - 1)

    # left_tps_i nao ha overap com left_tps_j, j != i, pq a card, eh differente, e a funcao *sempre* devolve uma arvore com card dada
    # left_tps_i nao pode haver repeticoes dentor, por inducao

    return rv


def err(msg):
    print("ERROR:", msg)
    exit(1)

if __name__ == "__main__":
    if len(sys.argv) != 4:
        err('USAGE: {} node_count depth cut-off'.format(sys.argv[0]))

    ls = enum2_eq_depth(*[int(p) for p in sys.argv[1:]])
    prns = []
    for t in ls:
        number_tree(t)
        t.print()
        print(": {} -- {}".format(depth_lb(t), depth_ub(t)))
        prns.append(t.to_string())

    print("topologies, ordered")
    prns.sort()
    for s in prns:
        print(s)
    for i in range(1, len(prns)):
        if prns[i] == prns[i-1]:
            print("duplicate {}", prns[i])


    # header, samples = run_sklearn.parse_samples(sys.stdin)
    # node_count, skl = run_sklearn.run_sk_learn(samples)
    # print ("# number of nodes:", node_count)
    # skl_topo = sklearn2topology(skl)
    # number_tree(skl_topo)
    # print ("# sklearn topology:")
    # skl_topo.print()
    # print()
    # print('-----')
    # show_nfolhas(skl_topo)
    # print('-----')
    # scored = [ (diff_topologies(skl_topo, t), i, t) for i, t in enumerate(ls) ]
    # scored.sort()
    # ls = [ (t, score) for score, _, t in scored ]
    # ln = len(ls)
    # if ln < 500:
    #     for t, score in ls:
    #         t.print()
    #         print(" diff: {}".format(score))
    # else:
    #     print("{:,}".format(ln))
